@@ -25,6 +25,8 @@
 #include <QModelIndex>
 #include <QProcess>
 #include <QDebug>
+#include <QQuickItem>
+#include <QQuickWindow>
 
 // KDE
 #include <KIO/ApplicationLauncherJob>
@@ -35,6 +37,11 @@
 #include <KSycoca>
 #include <KSycocaEntry>
 
+#include <KWayland/Client/connection_thread.h>
+#include <KWayland/Client/plasmawindowmanagement.h>
+#include <KWayland/Client/registry.h>
+#include <KWayland/Client/surface.h>
+
 constexpr int MAX_FAVOURITES = 5;
 
 ApplicationListModel::ApplicationListModel(HomeScreen *parent)
@@ -45,6 +52,7 @@ ApplicationListModel::ApplicationListModel(HomeScreen *parent)
             this, &ApplicationListModel::sycocaDbChanged);
 
     loadSettings();
+    initWayland();
 }
 
 ApplicationListModel::~ApplicationListModel() = default;
@@ -79,7 +87,8 @@ QHash<int, QByteArray> ApplicationListModel::roleNames() const
         {ApplicationEntryPathRole, QByteArrayLiteral("applicationEntryPath")},
         {ApplicationOriginalRowRole, QByteArrayLiteral("applicationOriginalRow")},
         {ApplicationStartupNotifyRole, QByteArrayLiteral("applicationStartupNotify")},
-        {ApplicationLocationRole, QByteArrayLiteral("applicationLocation")}
+        {ApplicationLocationRole, QByteArrayLiteral("applicationLocation")},
+        {ApplicationRunningRole, QByteArrayLiteral("applicationRunning")}
     };
 }
 
@@ -94,11 +103,62 @@ void ApplicationListModel::sycocaDbChanged(const QStringList &changes)
     loadApplications();
 }
 
-bool appNameLessThan(const ApplicationData &a1, const ApplicationData &a2)
+bool appNameLessThan(const ApplicationListModel::ApplicationData &a1, const ApplicationListModel::ApplicationData &a2)
 {
     return a1.name.toLower() < a2.name.toLower();
 }
 
+void ApplicationListModel::initWayland()
+{
+    if (!QGuiApplication::platformName().startsWith(QLatin1String("wayland"), Qt::CaseInsensitive)) {
+        return;
+    }
+    using namespace KWayland::Client;
+    ConnectionThread *connection = ConnectionThread::fromApplication(this);
+
+    if (!connection) {
+        return;
+    }
+    auto *registry = new Registry(this);
+    registry->create(connection);
+    connect(registry, &Registry::plasmaWindowManagementAnnounced, this,
+        [this, registry] (quint32 name, quint32 version) {
+            m_windowManagement = registry->createPlasmaWindowManagement(name, version, this);
+            qRegisterMetaType<QVector<int> >("QVector<int>");
+
+            connect(m_windowManagement, &KWayland::Client::PlasmaWindowManagement::windowCreated,
+                    this, [this] (KWayland::Client::PlasmaWindow *window) {
+                if (window->appId() == QStringLiteral("org.kde.plasmashell")) {
+                    return;
+                }
+                int idx = 0;
+                for (auto i = m_applicationList.begin(); i != m_applicationList.end(); i++) {
+                    if ((*i).storageId == window->appId() + QStringLiteral(".desktop")) {
+                        (*i).window = window;
+                        emit dataChanged(index(idx, 0), index(idx, 0));
+                        connect(window, &KWayland::Client::PlasmaWindow::unmapped, this, [this, window] () {
+                            int idx = 0;
+                            for (auto i = m_applicationList.begin(); i != m_applicationList.end(); i++) {
+                                if ((*i).storageId == window->appId() + QStringLiteral(".desktop")) {
+                                    (*i).window = nullptr;
+                                    emit dataChanged(index(idx, 0), index(idx, 0));
+                                    break;
+                                }
+                                idx++;
+                            }
+                        });
+                        break;
+                    }
+                    idx++;
+                }
+            });
+        }
+    );
+
+    registry->setup();
+    connection->roundtrip();
+}
+
 void ApplicationListModel::loadApplications()
 {
     auto cfg = KSharedConfig::openConfig(QStringLiteral("applications-blacklistrc"));
@@ -121,6 +181,7 @@ void ApplicationListModel::loadApplications()
 
     QMap<int, ApplicationData> orderedList;
     QList<ApplicationData> unorderedList;
+    QSet<QString> foundFavorites;
 
     // Iterate over all entries in the group
     while (!subGroupList.isEmpty()) {
@@ -159,6 +220,7 @@ void ApplicationListModel::loadApplications()
 
                             if (m_favorites.contains(data.storageId)) {
                                 data.location = Favorites;
+                                foundFavorites.insert(data.storageId);
                             } else if (m_desktopItems.contains(data.storageId)) {
                                 data.location = Desktop;
                             }
@@ -186,6 +248,18 @@ void ApplicationListModel::loadApplications()
 
     endResetModel();
     emit countChanged();
+
+    bool favChanged = false;
+    for (const auto &item : m_favorites) {
+        if (!foundFavorites.contains(item)) {
+            favChanged = true;
+            m_favorites.removeAll(item);
+        }
+    }
+    if (favChanged) {
+        m_homeScreen->config().writeEntry("Favorites", m_favorites);
+        emit favoriteCountChanged();
+    }
 }
 
 QVariant ApplicationListModel::data(const QModelIndex &index, int role) const
@@ -210,6 +284,8 @@ QVariant ApplicationListModel::data(const QModelIndex &index, int role) const
         return m_applicationList.at(index.row()).startupNotify;
     case ApplicationLocationRole:
         return m_applicationList.at(index.row()).location;
+    case ApplicationRunningRole:
+        return m_applicationList.at(index.row()).window != nullptr;
 
     default:
         return QVariant();
@@ -325,6 +401,13 @@ void ApplicationListModel::runApplication(const QString &storageId)
         return;
     }
 
+    for (auto i = m_applicationList.begin(); i != m_applicationList.end(); i++) {
+        if ((*i).window && (*i).storageId == storageId) {
+            (*i).window->requestActivate();
+            return;
+        }
+    }
+
     KService::Ptr service = KService::serviceByStorageId(storageId);
     KIO::ApplicationLauncherJob *job = new KIO::ApplicationLauncherJob(service);
     job->setUiDelegate(new KNotificationJobUiDelegate(KJobUiDelegate::AutoHandlingEnabled));
@@ -364,5 +447,61 @@ void ApplicationListModel::setMaxFavoriteCount(int count)
     emit maxFavoriteCountChanged();
 }
 
+void ApplicationListModel::setMinimizedDelegate(int row, QQuickItem *delegate)
+{
+    if (row < 0 || row >= m_applicationList.count()) {
+        return;
+    }
+
+    QWindow *delegateWindow = delegate->window();
+
+    if (!delegateWindow) {
+        return;
+    }
+
+    using namespace KWayland::Client;
+    KWayland::Client::PlasmaWindow *window = m_applicationList[row].window;
+    if (!window) {
+        return;
+    }
+
+    Surface *surface = Surface::fromWindow(delegateWindow);
+
+    if (!surface) {
+        return;
+    }
+
+    QRect rect = delegate->mapRectToScene(QRectF(0,0, delegate->width(), delegate->height())).toRect();
+
+    window->setMinimizedGeometry(surface, rect);
+}
+
+void ApplicationListModel::unsetMinimizedDelegate(int row, QQuickItem *delegate)
+{
+    if (row < 0 || row >= m_applicationList.count()) {
+        return;
+    }
+
+    QWindow *delegateWindow = delegate->window();
+
+    if (!delegateWindow) {
+        return;
+    }
+
+    using namespace KWayland::Client;
+    KWayland::Client::PlasmaWindow *window = m_applicationList[row].window;
+    if (!window) {
+        return;
+    }
+
+    Surface *surface = Surface::fromWindow(delegateWindow);
+
+    if (!surface) {
+        return;
+    }
+
+    window->unsetMinimizedGeometry(surface);
+}
+
 #include "moc_applicationlistmodel.cpp"
 
